@@ -13,7 +13,8 @@
 #    License for the specific language governing permissions and limitations
 #    under the License.
 #
-import six
+from datetime import timedelta
+import requests
 
 from gnocchiclient import auth as gauth
 from gnocchiclient import client as gclient
@@ -24,30 +25,19 @@ from oslo_log import log as logging
 from voluptuous import All
 from voluptuous import In
 from voluptuous import Length
+from voluptuous import Range
 from voluptuous import Required
 from voluptuous import Schema
 
 from cloudkitty import collector
+from cloudkitty.common import custom_session
+from cloudkitty import dataframe
 from cloudkitty import utils as ck_utils
-
+from cloudkitty.utils import tz as tzutils
 
 LOG = logging.getLogger(__name__)
 
-# NOTE(mc): The deprecated section should be removed in a future release.
 COLLECTOR_GNOCCHI_OPTS = 'collector_gnocchi'
-DEPRECATED_COLLECTOR_GNOCCHI_OPTS = 'gnocchi_collector'
-
-keystone_opts = ks_loading.get_auth_common_conf_options() + \
-    ks_loading.get_session_conf_options()
-
-keystone_opts = [
-    cfg.Opt(
-        o.name,
-        type=o.type,
-        help=o.help,
-        deprecated_group=DEPRECATED_COLLECTOR_GNOCCHI_OPTS,
-    ) for o in keystone_opts
-]
 
 collector_gnocchi_opts = [
     cfg.StrOpt(
@@ -56,39 +46,62 @@ collector_gnocchi_opts = [
         choices=['keystone', 'basic'],
         help='Gnocchi auth type (keystone or basic). Keystone credentials '
         'can be specified through the "auth_section" parameter',
-        deprecated_group=DEPRECATED_COLLECTOR_GNOCCHI_OPTS,
     ),
     cfg.StrOpt(
         'gnocchi_user',
         default='',
         help='Gnocchi user (for basic auth only)',
-        deprecated_group=DEPRECATED_COLLECTOR_GNOCCHI_OPTS,
     ),
     cfg.StrOpt(
         'gnocchi_endpoint',
         default='',
         help='Gnocchi endpoint (for basic auth only)',
-        deprecated_group=DEPRECATED_COLLECTOR_GNOCCHI_OPTS,
     ),
     cfg.StrOpt(
         'interface',
         default='internalURL',
         help='Endpoint URL type (for keystone auth only)',
-        deprecated_group=DEPRECATED_COLLECTOR_GNOCCHI_OPTS,
     ),
     cfg.StrOpt(
         'region_name',
         default='RegionOne',
         help='Region Name',
-        deprecated_group=DEPRECATED_COLLECTOR_GNOCCHI_OPTS,
     ),
+    cfg.IntOpt(
+        'http_pool_maxsize',
+        default=requests.adapters.DEFAULT_POOLSIZE,
+        help='If the value is not defined, we use the value defined by '
+             'requests.adapters.DEFAULT_POOLSIZE',
+    )
 ]
 
-cfg.CONF.register_opts(keystone_opts, COLLECTOR_GNOCCHI_OPTS)
+ks_loading.register_session_conf_options(cfg.CONF, COLLECTOR_GNOCCHI_OPTS)
+ks_loading.register_auth_conf_options(cfg.CONF, COLLECTOR_GNOCCHI_OPTS)
 cfg.CONF.register_opts(collector_gnocchi_opts, COLLECTOR_GNOCCHI_OPTS)
 
 CONF = cfg.CONF
 
+# According to 'gnocchi/rest/aggregates/operations.py#AGG_MAP' and
+# 'gnocchi/rest/aggregates/operations.py#AGG_MAP' the following are the basic
+# aggregation methods that one can use when configuring an aggregation
+# method in the archive policy in Gnocchi or using the aggregation API.
+BASIC_AGGREGATION_METHODS = set(('mean', 'sum', 'last', 'max', 'min', 'std',
+                                 'median', 'first', 'count'))
+for agg in list(BASIC_AGGREGATION_METHODS):
+    BASIC_AGGREGATION_METHODS.add("rate:%s" % agg)
+
+EXTRA_AGGREGATION_METHODS_FOR_ARCHIVE_POLICY = set(
+    (str(i) + 'pct' for i in range(1, 100)))
+
+for agg in list(EXTRA_AGGREGATION_METHODS_FOR_ARCHIVE_POLICY):
+    EXTRA_AGGREGATION_METHODS_FOR_ARCHIVE_POLICY.add("rate:%s" % agg)
+
+# The aggregation method that one can use to configure the archive
+# policies also supports the 'pct' (percentile) operation. Therefore,
+# we also expose this as a configuration.
+VALID_AGGREGATION_METHODS_FOR_METRICS = BASIC_AGGREGATION_METHODS.union(
+    EXTRA_AGGREGATION_METHODS_FOR_ARCHIVE_POLICY)
+
 GNOCCHI_EXTRA_SCHEMA = {
     Required('extra_args'): {
         Required('resource_type'): All(str, Length(min=1)),
@@ -96,7 +109,11 @@ GNOCCHI_EXTRA_SCHEMA = {
         # This parameter permits to adapt the key of the resource identifier
         Required('resource_key', default='id'): All(str, Length(min=1)),
         Required('aggregation_method', default='max'):
-            In(['max', 'mean', 'min', 'rate:max', 'rate:mean', 'rate:min']),
+            In(VALID_AGGREGATION_METHODS_FOR_METRICS),
+        Required('re_aggregation_method', default='max'):
+            In(BASIC_AGGREGATION_METHODS),
+        Required('force_granularity', default=3600): All(int, Range(min=0)),
+        Required('use_all_resource_revisions', default=True): All(bool),
     },
 }
 
@@ -115,8 +132,8 @@ class GnocchiCollector(collector.BaseCollector):
 
     collector_name = 'gnocchi'
 
-    def __init__(self, transformers, **kwargs):
-        super(GnocchiCollector, self).__init__(transformers, **kwargs)
+    def __init__(self, **kwargs):
+        super(GnocchiCollector, self).__init__(**kwargs)
 
         adapter_options = {'connect_retries': 3}
         if CONF.collector_gnocchi.gnocchi_auth_type == 'keystone':
@@ -140,7 +157,9 @@ class GnocchiCollector(collector.BaseCollector):
 
         self._conn = gclient.Client(
             '1',
-            session_options={'auth': auth_plugin, 'verify': verify},
+            session=custom_session.create_custom_session(
+                {'auth': auth_plugin, 'verify': verify},
+                CONF.collector_gnocchi.http_pool_maxsize),
             adapter_options=adapter_options,
         )
 
@@ -163,9 +182,8 @@ class GnocchiCollector(collector.BaseCollector):
         return output
 
     @classmethod
-    def get_metadata(cls, resource_name, transformers, conf):
-        info = super(GnocchiCollector, cls).get_metadata(resource_name,
-                                                         transformers)
+    def get_metadata(cls, resource_name, conf):
+        info = super(GnocchiCollector, cls).get_metadata(resource_name)
         try:
             info["metadata"].extend(
                 conf[resource_name]['groupby']
@@ -219,10 +237,10 @@ class GnocchiCollector(collector.BaseCollector):
         time_filter = list()
         time_filter.append(self.extend_filter(
             self.gen_filter(ended_at=None),
-            self.gen_filter(cop=">=", ended_at=start),
+            self.gen_filter(cop=">=", ended_at=start.isoformat()),
             lop='or'))
         time_filter.append(
-            self.gen_filter(cop="<=", started_at=end))
+            self.gen_filter(cop="<=", started_at=end.isoformat()))
         return time_filter
 
     def _fetch_resources(self, metric_name, start, end,
@@ -248,8 +266,9 @@ class GnocchiCollector(collector.BaseCollector):
         # FIXME(peschk_l): In order not to miss any resource whose metrics may
         # contain measures after its destruction, we scan resources over three
         # collect periods.
-        start -= CONF.collect.period
-        end += CONF.collect.period
+        delta = timedelta(seconds=CONF.collect.period)
+        start = tzutils.substract_delta(start, delta)
+        end = tzutils.add_delta(end, delta)
         query_parameters = self._generate_time_filter(start, end)
 
         if project_id:
@@ -278,7 +297,6 @@ class GnocchiCollector(collector.BaseCollector):
         """Get metric during the timeframe.
 
         :param metric_name: metric name to filter on.
-        :type resource_name: str
         :param start: Start of the timeframe.
         :param end: End of the timeframe if needed.
         :param project_id: Filter on a specific tenant/project.
@@ -286,50 +304,76 @@ class GnocchiCollector(collector.BaseCollector):
         :param q_filter: Append a custom filter.
         :type q_filter: list
         """
+        agg_kwargs = self.get_aggregation_api_arguments(end, metric_name,
+                                                        project_id,
+                                                        q_filter,
+                                                        start)
 
-        # Get gnocchi specific conf
-        extra_args = self.conf[metric_name]['extra_args']
-
-        # get ressource type
-        resource_type = extra_args['resource_type']
-        scope_key = CONF.collect.scope_key
-
-        # build search query using ressource type and project_id if provided
-        query_parameters = list()
-        query_parameters.append(
-            self.gen_filter(cop="=", type=resource_type))
-
-        if project_id:
-            kwargs = {scope_key: project_id}
-            query_parameters.append(self.gen_filter(**kwargs))
-        if q_filter:
-            query_parameters.append(q_filter)
-
-        # build aggregration operation
-        op = ["aggregate", extra_args['aggregation_method'],
-              ["metric", metric_name, extra_args['aggregation_method']]]
-
-        # get groupby
-        groupby = self.conf[metric_name]['groupby']
-
+        op = self.build_operation_command(metric_name)
         try:
-            return self._conn.aggregates.fetch(
-                op,
-                resource_type=resource_type,
-                start=ck_utils.ts2dt(start),
-                stop=ck_utils.ts2dt(end),
-                groupby=groupby,
-                search=self.extend_filter(*query_parameters))
+            measurements = self._conn.aggregates.fetch(op, **agg_kwargs)
+            LOG.debug("Measurements [%s] received with operation [%s] and "
+                      "arguments [%s].", measurements, op, agg_kwargs)
+            return measurements
         except (gexceptions.MetricNotFound, gexceptions.BadRequest) as e:
             # FIXME(peschk_l): gnocchiclient seems to be raising a BadRequest
             # when it should be raising MetricNotFound
             if isinstance(e, gexceptions.BadRequest):
-                if 'Metrics not found' not in six.text_type(e):
+                if 'Metrics not found' not in e.args[0]:
                     raise
             LOG.warning('[{scope}] Skipping this metric for the '
                         'current cycle.'.format(scope=project_id, err=e))
             return []
 
+    def get_aggregation_api_arguments(self, end, metric_name, project_id,
+                                      q_filter, start):
+        extra_args = self.conf[metric_name]['extra_args']
+        resource_type = extra_args['resource_type']
+
+        query_parameters = self.build_query_parameters(project_id, q_filter,
+                                                       resource_type)
+        agg_kwargs = {
+            'resource_type': resource_type,
+            'start': start,
+            'stop': end,
+            'groupby': self.conf[metric_name]['groupby'],
+            'search': self.extend_filter(*query_parameters),
+        }
+
+        force_granularity = extra_args['force_granularity']
+        if force_granularity > 0:
+            agg_kwargs['granularity'] = force_granularity
+
+        re_aggregation_method = extra_args['re_aggregation_method']
+        if re_aggregation_method.startswith('rate:'):
+            agg_kwargs['start'] = start - timedelta(seconds=force_granularity)
+            LOG.debug("Re-aggregation method for metric [%s] configured as"
+                      " [%s]. Therefore, we need two data points. Start date"
+                      " modified from [%s] to [%s].", metric_name,
+                      re_aggregation_method, start, agg_kwargs['start'])
+
+        return agg_kwargs
+
+    def build_query_parameters(self, project_id, q_filter, resource_type):
+        query_parameters = list()
+        query_parameters.append(
+            self.gen_filter(cop="=", type=resource_type))
+        if project_id:
+            scope_key = CONF.collect.scope_key
+            kwargs = {scope_key: project_id}
+            query_parameters.append(self.gen_filter(**kwargs))
+        if q_filter:
+            query_parameters.append(q_filter)
+        return query_parameters
+
+    def build_operation_command(self, metric_name):
+        extra_args = self.conf[metric_name]['extra_args']
+
+        re_aggregation_method = extra_args['re_aggregation_method']
+        op = ["aggregate", re_aggregation_method,
+              ["metric", metric_name, extra_args['aggregation_method']]]
+        return op
+
     def _format_data(self, metconf, data, resources_info=None):
         """Formats gnocchi data to CK data.
 
@@ -368,6 +412,9 @@ class GnocchiCollector(collector.BaseCollector):
             q_filter=q_filter,
         )
 
+        data = GnocchiCollector.filter_unecessary_measurements(
+            data, met, metric_name)
+
         resources_info = None
         if met['metadata']:
             resources_info = self._fetch_resources(
@@ -377,9 +424,13 @@ class GnocchiCollector(collector.BaseCollector):
                 project_id=project_id,
                 q_filter=q_filter
             )
+
         formated_resources = list()
         for d in data:
             # Only if aggregates have been found
+            LOG.debug("Processing entry [%s] for [%s] in timestamp ["
+                      "start=%s, end=%s] and project id [%s]", d,
+                      metric_name, start, end, project_id)
             if d['measures']['measures']['aggregated']:
                 try:
                     metadata, groupby, qty = self._format_data(
@@ -388,17 +439,51 @@ class GnocchiCollector(collector.BaseCollector):
                     LOG.warning(
                         '[{}] An error occured during data collection '
                         'between {} and {}: {}'.format(
-                            project_id,
-                            ck_utils.ts2dt(start),
-                            ck_utils.ts2dt(end),
-                            e),
+                            project_id, start, end, e),
                     )
                     continue
-                data = self.t_cloudkitty.format_item(
+                formated_resources.append(dataframe.DataPoint(
+                    met['unit'],
+                    qty,
+                    0,
                     groupby,
                     metadata,
-                    met['unit'],
-                    qty=qty,
-                )
-                formated_resources.append(data)
+                ))
         return formated_resources
+
+    @staticmethod
+    def filter_unecessary_measurements(data, met, metric_name):
+        """Filter unecessary measurements if not 'use_all_resource_revisions'
+
+        The option 'use_all_resource_revisions' is useful when using Gnocchi
+        with the patch introduced in
+        https://github.com/gnocchixyz/gnocchi/pull/1059.
+
+        That patch can cause queries to return more than one entry per
+        granularity (timespan), according to the revisions a resource has.
+        This can be problematic when using the 'mutate' option of Cloudkitty.
+        Therefore, this option ('use_all_resource_revisions') allows operators
+        to discard all datapoints returned from Gnocchi, but the last one in
+        the granularity queried by CloudKitty. The default behavior is
+        maintained, which means, CloudKitty always use all of the data
+        points returned.
+        """
+
+        use_all_resource_revisions = \
+            met['extra_args']['use_all_resource_revisions']
+        LOG.debug("Configuration use_all_resource_revisions set to [%s] for "
+                  "%s", use_all_resource_revisions, metric_name)
+
+        if data and not use_all_resource_revisions:
+            data.sort(
+                key=lambda x: (x["group"]["id"], x["group"]["revision_start"]),
+                reverse=False)
+
+            # We just care about the oldest entry per resource ID in the
+            # given time slice (configured granularity in Cloudkitty).
+            single_entries_per_id = {d["group"]["id"]: d for d in
+                                     data}.values()
+            LOG.debug("Replaced list of data points [%s] with [%s] for "
+                      "metric [%s]", data, single_entries_per_id, metric_name)
+            data = single_entries_per_id
+        return data
